from conftest import EMPTY_WANT, compare_have_want


def test_no_manuscript_title():
    compare_have_want(
        have="""\
        Lorem ipsum.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Lorem ipsum.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_manuscript_title():
    compare_have_want(
        have="""\
        # My Title

        Lorem ipsum.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="paragraph" data-nodeid="1">

        <p>Lorem ipsum.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_manuscript_with_shortcut_title():
    compare_have_want(
        have="""\
        # Title
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>Title</h1>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_manuscript_meta():
    compare_have_want(
        have="""\
        # My Title {
          :label: mylbl
          :title: My Title
          :date: 2022-03-29
        }
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div id="mylbl" class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <p class="manuscript-date">March 29, 2022</p>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_section_header():
    compare_have_want(
        have="""\
        Lorem ipsum.

        ## section title

        Lorem ipsum.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Lorem ipsum.</p>

        </div>

        <section class="section level-2" data-nodeid="3">

        <h2>1. section title</h2>

        <div class="paragraph" data-nodeid="4">

        <p>Lorem ipsum.</p>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )



"""Nodes that make up the manuscript tree.

The manuscript tree semantically represents each element in the manuscript such as
sections, paragraphs, special regions, figures, etc.

This is opposed to the syntax tree output by tree-sitter whose nodes represent elements
with syntactic meaning such as tags, delimiters, etc.

There are two main classes of interest: :class:`Node` and :class:`NodeWithChildren`.
They implement 95% of the entire API of this module.  Other classes are mostly dummy
subclasses whose only purpose is to differentiate themselves from others via their type.

"""

import logging
import textwrap
from collections.abc import Callable, Generator, Iterable
from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    ClassVar,
    Optional,
    TypeVar,
    Union,
    cast,
)

logger = logging.getLogger("RSM").getChild("nodes")

NodeSubType = TypeVar("NodeSubType", bound="Node")


class RSMNodeError(Exception):
    pass


class Node:
    """A node in the manuscript tree.

    A node represents a semantically meaningful element of the manuscript.

    Parameters
    ----------
    label
        Unique identifier for this node.
    types
        Types of this node.
    number
        Node number.
    nonum
        Whether this node should be automatically given a number.
    reftext_template
        If not empty, replaces :attr:`classreftext`.

    See Also
    --------
    :class:`Manuscript` : The class of the root node of a manuscript tree.
    :class:`NodeWithChildren` : Subclass that implements methods to handle children.

    Notes
    -----
    An instance of Node cannot have children.  Only instances of (subclasses) of
    :class:`NodeWithChildren` may have them.  However, for the sake of having a uniform
    API, Node implements the property :attr:`children`, which always returns an empty
    tuple.

    """

    classreftext: ClassVar[str] = "{nodeclass} {number}"

    possible_parents: ClassVar[set[type["NodeWithChildren"]]] = set()
    """Allowed types of parent Nodes.

    When setting the parent of a Node, this attribute is checked to see whether the
    intended parent has an admissible type.

    Examples
    --------
    This is a class variable

    >>> nodes.Item.possible_parents == {nodes.Itemize, nodes.Enumerate, nodes.Contents}
    True

    This variable is checked when setting the parent directly.

    >>> it = nodes.Item()
    >>> it.parent = nodes.Paragraph()
    Traceback (most recent call last):
    rsm.nodes.RSMNodeError: Node of type <class 'rsm.nodes.Item'> cannot have parent of type <class 'rsm.nodes.Paragraph'>

    This is also used when setting the parent in some other indirect way, for example
    when calling :meth:`~NodeWithChildren.append` on the desired parent.

    >>> nodes.Paragraph().append(it)
    Traceback (most recent call last):
    rsm.nodes.RSMNodeError: Node of type <class 'rsm.nodes.Item'> cannot have parent of type <class 'rsm.nodes.Paragraph'>

    Allowed parents proceed without raising.

    >>> nodes.Itemize().append(it)
    Itemize(parent=None, [Item])

    """

    autonumber: ClassVar[bool] = False
    """Whether to automatically assign a number to this node during transform step.

    Examples
    --------
    >>> msc, thm1, thm2 = nodes.Manuscript(), nodes.Theorem(), nodes.Theorem()
    >>> (thm1.number, thm2.number) == (None, None)
    True
    >>> tform = rsm.transformer.Transformer()
    >>> tform.transform(msc.append([thm1, thm2]))  # doctest: +IGNORE_RESULT
    >>> thm1.number, thm2.number
    (1, 2)

    """

    _number_within: ClassVar[type["Node"] | None] = None
    # see property number_within for documentation

    _number_as: ClassVar[type["Node"] | None] = None
    # see property number_as for documentation

    newmetakeys: ClassVar[set] = {"label", "types", "nonum", "reftext"}
    """Meta keys to add to those of the parent class.

    .. important::
       Only use this when defining a new Node subclass.  When dealing with Node
       isntances, do not access this attribute directly neither for reading nor writing.
       Always use :meth:`metakeys` in that case.

    See Also
    --------
    :meth:`metakeys`

    Examples
    --------
    The keys in `newmetakeys` are added to the meta keys of the parent class.

    >>> nodes.Heading.newmetakeys
    {'title'}
    >>> nodes.Heading.metakeys() == nodes.Node.metakeys() | {"title"}
    True

    The intended use, and only supported use, of `newmetakeys` is at the time of class
    definition.

    >>> class NewNode(nodes.Node):
    ...     newmetakeys = {"newkey"}
    >>> NewNode.metakeys() == nodes.Node.metakeys() | {"newkey"}
    True

    """

    has_handrail: ClassVar[bool] = False
    """Whether nodes of this class are rendered with a handrail."""

    def __init__(
        self,
        label: str = "",
        types: list[str] | None = None,
        number: int | None = None,
        nonum: bool = False,
        reftext_template: str = "",
        start_point: tuple[int, int] = (-1, -1),
        end_point: tuple[int, int] = (-1, -1),
    ) -> None:
        self.nodeid: str | None = None
        """Node id - always exists (unlike label), automatically assigned, unique within the tree."""
        self.label: str = label
        """Unique identifier."""
        self.types: list[str] = types or []
        """Types of this node."""
        self.handrail_depth = 0
        """The number of ancestors of this node that have a handrail."""
        self.number: int | None = number
        """Node number."""
        self.nonum: bool = nonum
        """Whether this node should be automatically given a number."""
        self.reftext_template: str = reftext_template or self.classreftext
        """Reftext template, or "" to use :attr:`classreftext`."""
        self.start_point: tuple[int, int] = start_point
        """The start point of the corresponding concrete syntax tree node."""
        self.end_point: tuple[int, int] = end_point
        """The end point of the corresponding concrete syntax tree node."""
        self._parent: NodeWithChildren | None = None

    def _attrs_for_repr_and_eq(self) -> list[str]:
        return ["label", "types", "nonum", "number", "parent"]

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        d = {
            att: getattr(self, att)
            for att in self._attrs_for_repr_and_eq()
            if att != "children"
        }
        d["parent"] = (
            "None" if self.parent is None else f"{self.parent.__class__.__name__}"
        )
        d_str = ", ".join(f"{k}={v}" for k, v in d.items() if v)
        return f"{cls}({d_str})"

    def __eq__(self, other: Any) -> bool:
        attrs = self._attrs_for_repr_and_eq()
        try:
            return all(
                (mine is getattr(other, a))
                if isinstance(mine := getattr(self, a), Node)
                else (mine == getattr(other, a))
                for a in attrs
            )
        except AttributeError:
            return False

    def sexp(
        self,
        tab_width: int = 2,
        meta: bool = False,
        ignore_meta_keys: set | None = None,
    ) -> str:
        """Represent this node as an S expression.

        Parameters
        ----------
        tab_width
            How many spaces of indentation at each depth level.
        meta
            Whether to include meta in the output.
        ignore_meta_keys
            When `meta` is ``True``, this controls which meta keys to include.

        Returns
        -------
        A string representation of the Node and its descendents.

        See Also
        --------
        :meth:`traverse`

        Examples
        --------
        >>> p1, p2, p3, p4 = [nodes.Paragraph().append(nodes.Text()) for _ in range(4)]
        >>> msc = nodes.Manuscript().append(
        ...     [
        ...         nodes.Section().append(p1),
        ...         nodes.Section().append([nodes.Subsection().append([p2, p3])]),
        ...         nodes.Section().append(p4),
        ...     ]
        ... )

        By default, meta keys are not included in the output.

        >>> print(msc.sexp())
        (Manuscript
          (Section
            (Paragraph
              (Text)))
          (Section
            (Subsection
              (Paragraph
                (Text))
              (Paragraph
                (Text))))
          (Section
            (Paragraph
              (Text))))

        :meth:`sexp` is useful when transforming the tree.

        >>> p3.replace_self(nodes.Paragraph().append(nodes.Span(strong=True).append(nodes.Text())))
        >>> print(msc.sexp())
        (Manuscript
          (Section
            (Paragraph
              (Text)))
          (Section
            (Subsection
              (Paragraph
                (Text))
              (Paragraph
                (Span
                  (Text)))))
          (Section
            (Paragraph
              (Text))))

        Output meta for even more debugging information.

        >>> print(msc.sexp(meta=True))
        (Manuscript { :reftext: Manuscript }
          (Section { :reftext: Section }
            (Paragraph { :reftext: Paragraph }
              (Text { :reftext: Text })))
          (Section { :reftext: Section }
            (Subsection { :reftext: Section }
              (Paragraph { :reftext: Paragraph }
                (Text { :reftext: Text }))
              (Paragraph { :reftext: Paragraph }
                (Span { :reftext: Span, :strong: True }
                  (Text { :reftext: Text })))))
          (Section { :reftext: Section }
            (Paragraph { :reftext: Paragraph }
              (Text { :reftext: Text }))))

        Use `ignore_meta_keys` for a less verbose output.

        >>> print(msc.sexp(meta=True, ignore_meta_keys={"reftext"}))
        (Manuscript {  }
          (Section {  }
            (Paragraph {  }
              (Text {  })))
          (Section {  }
            (Subsection {  }
              (Paragraph {  }
                (Text {  }))
              (Paragraph {  }
                (Span { :strong: True }
                  (Text {  })))))
          (Section {  }
            (Paragraph {  }
              (Text {  }))))

        """
        ignore_meta_keys = set() if ignore_meta_keys is None else set(ignore_meta_keys)
        exp = ""
        stack = [(0, self)]
        while stack:
            indent, node = stack.pop()
            if node is None:
                exp += ")"
                continue
            spaces = " " * indent
            exp += f"\n{spaces}({node.__class__.__name__}"
            if meta:
                meta_str = (
                    "{ "
                    + ", ".join(
                        [
                            f":{key}: {val}"
                            for key in sorted(node.metakeys())
                            if key not in ignore_meta_keys
                            and (val := getattr(node, key))
                        ]
                    )
                    + " }"
                )
                exp += f" {meta_str}"
            stack.append((None, None))
            if node.children:
                stack += [(indent + tab_width, c) for c in reversed(node.children)]
        return exp[1:]  # get rid of an extra newline at the start

    @classmethod
    def metakeys(cls: type["Node"]) -> set[str]:
        """The valid meta keys of the given class.

        Returns
        -------
        The valid meta keys.

        Examples
        --------
        >>> all_nodes_meta = {"label", "types", "nonum", "reftext"}
        >>> nodes.Node.metakeys() == all_nodes_meta
        True
        >>> nodes.Span.metakeys() - all_nodes_meta == {"strong", "emphas", "little", "insert", "delete"}
        True
        >>> nodes.Author.metakeys()  - all_nodes_meta == {"name", "affiliation", "email", "orcid", "author_note"}
        True

        """
        return cls.newmetakeys.union(
            *[b.metakeys() for b in cls.__bases__ if hasattr(b, "metakeys")]
        )

    @property
    def parent(self) -> Optional["NodeWithChildren"]:
        """The parent of this Node, or None."""
        return self._parent

    @parent.setter
    def parent(self, node: Optional["NodeWithChildren"]) -> None:
        if node is None:
            self._parent = None
        elif not self.__class__.possible_parents:
            self._parent = node
        else:
            possible_parents = self.__class__.possible_parents
            if possible_parents and type(node) not in possible_parents:
                raise RSMNodeError(
                    f"Node of type {type(self)} cannot have parent of type {type(node)}"
                )
            self._parent = node

    @property
    def children(self) -> tuple:
        """Tuple with this Node's children."""
        return ()  # necessary for methods such as Nodes.traverse

    @property
    def number_within(self) -> type["Node"]:
        return self.__class__._number_within or Manuscript

    @property
    def number_as(self) -> type["Node"]:
        return self._number_as or self.__class__

    @property
    def full_number(self) -> str | None:
        if self.nonum:
            return None
        ancestor = self.first_ancestor_of_type(self.number_within)
        if not ancestor:
            logger.warning(
                f"{self.__class__.__name__} numbered within "
                f"{self.number_within} but no such ancestor was found; "
                "using root node instead"
            )
            ancestor = self.first_ancestor_of_type(Manuscript)
        if ancestor and ancestor.full_number:
            return f"{ancestor.full_number}.{self.number}"
        return f"{self.number}" if self.number else ""

    @property
    def reftext(self) -> str:
        return self.reftext_template.format(
            nodeclass=self.__class__.__name__, number=self.full_number
        ).strip()

    def traverse(
        self,
        *,
        condition: Callable[["Node"], bool] | None = None,
        nodeclass: type["Node"] | None = None,
    ) -> Generator[NodeSubType, None, None]:
        """Generate the descendents of this Node in depth-first order.

        By default, yield this node and then every descendent in depth-first order.  If
        `condition` is given, yield only those Nodes that satisfy the condition.  If
        `nodeclass` is given, it overrides `condition` and only those descendents of the
        specified type are yielded.

        Parameters
        ----------
        condition
            Callable that receives a single argument, a descendent Node, and returns
            whether it should be yielded.
        nodeclass
            A Node subclass of the desired yielded descendent Nodes.  If not None,
            `condition` is ignored.

        Yields
        -------
        :class:`Node`

        See Also
        --------
        :meth:`sexp`

        Notes
        -----
        Passing ``nodeclass=<NodeSubType>`` is equivalent to passing ``condition=lambda
        n: isinstance(n, <NodeSubType>)``.

        Examples
        --------
        >>> p1, p2, p3, p4 = [nodes.Paragraph().append(nodes.Text()) for _ in range(4)]
        >>> msc = nodes.Manuscript().append(
        ...     [
        ...         nodes.Section().append(p1),
        ...         nodes.Section().append([nodes.Subsection().append([p2, p3])]),
        ...         nodes.Section().append(p4),
        ...     ]
        ... )

        Visit every descendent, including self.

        >>> for n in msc.traverse(): print(n)
        Manuscript(parent=None, [Section, Section, Section])
        Section(parent=Manuscript, [Paragraph])
        Paragraph(parent=Section, [Text])
        Text("")
        Section(parent=Manuscript, [Subsection])
        Subsection(parent=Section, [Paragraph, Paragraph])
        Paragraph(parent=Subsection, [Text])
        Text("")
        Paragraph(parent=Subsection, [Text])
        Text("")
        Section(parent=Manuscript, [Paragraph])
        Paragraph(parent=Section, [Text])
        Text("")

        Use `nodeclass` to yield only nodes of a specified type.  Note that subclasses
        are also yielded.

        >>> for n in msc.traverse(nodeclass=nodes.Section): print(n)
        Section(parent=Manuscript, [Paragraph])
        Section(parent=Manuscript, [Subsection])
        Subsection(parent=Section, [Paragraph, Paragraph])
        Section(parent=Manuscript, [Paragraph])

        Yield only nodes satisfying an arbitrary condition

        >>> msc.children[1].nonum = True
        >>> for n in msc.traverse(condition=lambda n: n.nonum): print(n)
        Section(nonum=True, parent=Manuscript, [Subsection])

        """
        if condition is None:

            def condition(n):
                return True

        if nodeclass is not None:
            if issubclass(nodeclass, Node):

                def condition(n):
                    return isinstance(n, nodeclass)
            else:
                raise RSMNodeError("nodeclass must inherit from Node")

        stack = [self]
        while stack:
            node = stack.pop()
            if condition(node):
                yield cast(NodeSubType, node)
            stack += node.children[::-1]

    def first_of_type(
        self, cls: type["Node"] | tuple[type["Node"]], return_idx: bool = False
    ) -> Optional["Node"] | tuple[Optional["Node"], int | None]:
        """First child of the specified type.

        Parameters
        ----------
        cls
            Desired class of the child.
        return_idx
            Whether to return the index of the child among this node's children.

        Returns
        -------
        Node
            The first child of the specified type, or None.
        Node, int
            If `return_idx` is True, return (child, index), or (None, None).


        See Also
        --------
        :meth:`last_of_type`
        :meth:`first_ancestor_of_type`

        Examples
        --------
        >>> p = nodes.Paragraph().append([nodes.Text("one"), nodes.Text("two")])
        >>> p.first_of_type(nodes.Text)
        Text("one")
        >>> p.first_of_type(nodes.Text, return_idx=True)
        (Text("one"), 0)

        The index counts all existing children.

        >>> p.prepend(nodes.Span())
        Paragraph(parent=None, [Span, Text, Text])
        >>> p.first_of_type(nodes.Text, return_idx=True)
        (Text("one"), 1)

        """
        for idx, child in enumerate(self.children):
            if isinstance(child, cls):
                return (child, idx) if return_idx else child
        return (None, None) if return_idx else None

    def last_of_type(
        self, cls: type["Node"] | tuple[type["Node"]], return_idx: bool = False
    ) -> Optional["Node"] | tuple[Optional["Node"], int | None]:
        """Last child of the specified type.

        For details, see :meth:`first_of_type`.

        Examples
        --------
        >>> p = nodes.Paragraph().append([nodes.Span(), nodes.Text("one"), nodes.Text("two")])
        >>> p.last_of_type(nodes.Text, return_idx=True)
        (Text("two"), 2)

        """
        for idx, child in enumerate(reversed(self.children)):
            if isinstance(child, cls):
                return (child, len(self.children) - idx - 1) if return_idx else child
        return (None, None) if return_idx else None

    def prev_sibling(self, cls: type["Node"] | str | None = None) -> Optional["Node"]:
        """The previous sibling, optionally of a specified type.

        Parameters
        ----------
        cls
            The type of the desired sibling.  If ``"self"``, search for the previous
            sibling with the same type as this node.  If ``None``, return the
            immediately preceding sibling, regardless of its type.

        Returns
        -------
            The desired sibling, or None.

        See Also
        --------
        :meth:`first_ancestor_of_type`
        :meth:`next_sibling`

        Examples
        --------
        >>> p, s, t1, t2 = nodes.Paragraph(), nodes.Span(), nodes.Text("one"), nodes.Text("two")
        >>> p.append([t1, s, t2])  # doctest: +IGNORE_RESULT
        >>> t2.prev_sibling()
        Span(parent=Paragraph)
        >>> t2.prev_sibling(nodes.Text)
        Text("one")
        >>> t1.prev_sibling() is None
        True

        Use ``"self"`` to find nodes of the same type.

        >>> s2 = nodes.Span()
        >>> p.append(s2)
        Paragraph(parent=None, [Text, Span, Text, Span])
        >>> s2.prev_sibling() is t2
        True
        >>> s2.prev_sibling("self") is s
        True

        """
        if self.parent is None:
            return None

        ids = [id(c) for c in self.parent.children]
        index = ids.index(id(self))

        if cls is None and index:
            return self.parent.children[index - 1]

        if cls == "self":
            cls = self.__class__
        cls = cast(type["Node"], cls)

        prev_sibs = self.parent.children[:index]
        for node in reversed(prev_sibs):
            if isinstance(node, cls):
                return node
        return None

    def next_sibling(self, cls: type["Node"] | None = None) -> Optional["Node"]:
        """The next sibling, optionally of a specified type.

        For details, see :meth:`prev_sibling`.

        See Also
        --------
        :meth:`prev_sibling`

        """
        if self.parent is None:
            return None

        ids = [id(c) for c in self.parent.children]
        index = ids.index(id(self))

        if cls is None and index < len(self.parent.children) - 1:
            return self.parent.children[index + 1]

        if cls == "self":
            cls = self.__class__
        cls = cast(type["Node"], cls)

        next_sibs = self.parent.children[index + 1 :]
        for node in next_sibs:
            if isinstance(node, cls):
                return node
        return None

    def first_ancestor_of_type(
        self, cls: type["Node"] | tuple[type["Node"]]
    ) -> Optional["Node"]:
        """First ancestor of the specified type.

        Parameters
        ----------
        cls
            Desired class of the ancestor.

        Returns
        -------
        The first ancestor of the specified type, or None.

        See Also
        --------
        :attr:`parent`
        :meth:`first_of_type`

        Examples
        --------
        Given the tree

        >>> t = nodes.Text("Hello")
        >>> p = nodes.Paragraph().append(nodes.Span().append(t))
        >>> print(p.sexp())
        (Paragraph
          (Span
            (Text)))

        Find an ancestor of a desired type.

        >>> t.first_ancestor_of_type(nodes.Paragraph)
        Paragraph(parent=None, [Span])

        Always check the return value against ``None``.

        >>> t.first_ancestor_of_type(nodes.Manuscript) is None
        True

        """
        ancestor = self.parent
        # We use `type is not cls` instead of the recommended `isinstance()` because we
        # are looking for an exact type, not a subtype.  For example, we may want to
        # find the enclosing Section of a Theorem, bypassing any Subsections that may
        # lie in between.
        while ancestor and (
            all(type(ancestor) is not c for c in cls)
            if isinstance(cls, tuple)
            else (type(ancestor) is not cls)
        ):
            ancestor = ancestor.parent
        return ancestor  # the root node has parent None

    def get_child_by_id(self, nodeid: int) -> Optional["Node"]:
        for node in self.traverse(condition=lambda n: n.nodeid == nodeid):
            return node
        else:
            return None

    def replace_self(self, replacement: Union["Node", Iterable["Node"]]) -> None:
        """Replace this node in its parent's children.

        This is mostly useful during the transform step.

        Parameters
        ----------
        replacement
            The Node or Nodes to replace this with.

        Raises
        ------
        RSMNodeError
            If this Node's parent is None.

        See Also
        --------
        :meth:`remove_self`

        Examples
        --------
        Wrap a Text in a strong Span to render it in bold face.

        >>> p, t = nodes.Paragraph(), nodes.Text("one")
        >>> p.append(t)         # doctest: +IGNORE_RESULT
        >>> print(p.sexp())
        (Paragraph
          (Text))
        >>> s = nodes.Span(strong=True)
        >>> t.replace_self(s)
        >>> s.append(t)         # doctest: +IGNORE_RESULT
        >>> print(p.sexp())
        (Paragraph
          (Span
            (Text)))

        Note the call to :meth:`replace_self` must happen *before* the Text is added to
        the Span.

        May also replace with a list of Nodes.

        >>> t.replace_self([nodes.Text("new one"), nodes.Text("two")])
        >>> print(p.sexp())
        (Paragraph
          (Span
            (Text)
            (Text)))

        The following recipe uses the above example to wrap every Text within ``root``
        in a strong Span.  Note this is done to each Text descendent of ``root``,
        regardless of depth, and without any reference to their original immediate
        parents.

        >>> root = nodes.Manuscript().append(...)    # doctest: +SKIP
        >>> for t in root.traverse(nodeclass=Text):  # doctest: +SKIP
        ...     s = nodes.Span(strong=True)
        ...     t.replace_self(s)
        ...     s.append(t)

        """
        if not self.parent:
            raise RSMNodeError("Can only call replace_self on a node with parent")
        ids = [id(c) for c in self.parent.children]
        index = ids.index(id(self))
        parent = self.parent
        self.parent.remove(self)
        if not isinstance(replacement, Node):
            for idx, rep in enumerate(replacement):
                parent._children.insert(index + idx, rep)
                rep.parent = parent
        else:
            parent._children.insert(index, replacement)
            replacement.parent = parent

    def remove_self(self) -> None:
        """Remove this Node from its parent's children.

        See Also
        --------
        :meth:`replace_self`

        Examples
        --------
        >>> t = nodes.Text("remove me")
        >>> p = nodes.Paragraph().append(t)
        >>> p.children
        (Text("remove me"),)
        >>> t.remove_self()
        >>> p.children
        ()

        """
        if self.parent is not None:
            self.parent.remove(self)
            self.parent = None

    def ingest_dict_as_meta(self, meta: dict) -> None:
        if "reftext" in meta:
            meta["reftext_template"] = meta["reftext"]
            del meta["reftext"]
        for key, value in meta.items():
            setattr(self, str(key), value)


class NodeWithChildren(Node):
    """A :class:`Node` that may have children Nodes."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._children: list[Node] = []

    def _attrs_for_repr_and_eq(self) -> list[str]:
        return super()._attrs_for_repr_and_eq() + ["children"]

    def __repr__(self) -> str:
        if not self._children:
            return super().__repr__()
        children_repr = ", ".join(f"{c.__class__.__name__}" for c in self._children)
        children_repr = "[" + children_repr + "]"
        ret = super().__repr__()
        return ret[:-1] + ", " + children_repr + ")"

    @property
    def children(self) -> tuple[Node, ...]:
        """Tuple with this Node's children."""
        return tuple(self._children)

    def clear(self) -> None:
        """Remove all children."""
        for c in self._children:
            c.parent = None
        self._children = []

    def append(self, child: Node | Iterable[Node]) -> "NodeWithChildren":
        """Add a child or children after all current children.

        Parameters
        ----------
        child
            Node or iterable or nodes to append.

        Returns
        -------
        self

        Raises
        ------
        RSMNodeError
            When attempting to add as a child a node that already has a parent.

        See Also
        --------
        :meth:`prepend`

        Examples
        --------

        Append one or many children.

        >>> p, t1, t2, t3 = nodes.Paragraph(), nodes.Text("one"), nodes.Text("two"), nodes.Text("three")
        >>> p.append(t1)        # doctest: +IGNORE_RESULT
        >>> p.append([t2, t3])  # doctest: +IGNORE_RESULT
        >>> p.children
        (Text("one"), Text("two"), Text("three"))

        Calls can be chained.

        >>> p, t1, t2, t3 = nodes.Paragraph(), nodes.Text("one"), nodes.Text("two"), nodes.Text("three")
        >>> p.append(t1).append([t2, t3])  # doctest: +IGNORE_RESULT
        >>> p.children
        (Text("one"), Text("two"), Text("three"))

        Will raise when trying to append a node that already has a parent.

        >>> p2 = nodes.Paragraph()
        >>> p2.append(t1)
        Traceback (most recent call last):
        rsm.nodes.RSMNodeError: Attempting to append child to a new parent

        This is the case even when appending to the same parent.

        >>> t4 = nodes.Text("four")
        >>> p2.append(t4)       # doctest: +IGNORE_RESULT
        >>> p2.append(t4)
        Traceback (most recent call last):
        rsm.nodes.RSMNodeError: Attempting to append child to a new parent

        """
        if isinstance(child, Iterable):
            for c in child:
                self.append(c)
        elif isinstance(child, Node):
            if child.parent:
                raise RSMNodeError("Attempting to append child to a new parent")
            self._children.append(child)
            child.parent = self
        else:
            raise TypeError("Can only append a Node or iterable of Nodes as children")
        return self

    def prepend(self, child: Node | Iterable[Node]) -> "NodeWithChildren":
        """Add a child or children before all current children.

        For details, see :meth:`append`.

        See Also
        --------
        :meth:`append`

        """
        if isinstance(child, list):
            for c in reversed(child):
                self.prepend(c)
        elif isinstance(child, Node):
            if child.parent and child.parent is not self:
                raise RSMNodeError("Attempting to prepend child to a different parent")
            self._children.insert(0, child)
            child.parent = self
        else:
            raise TypeError("Can only prepend a Node or iterable of Nodes as children")
        return self

    def remove(self, child: "Node") -> None:
        """Remove child."""
        ids = [id(c) for c in self._children]
        index = ids.index(id(child))
        del self._children[index]
        child.parent = None


class Text(Node):
    """Plain text node.  Cannot contain children."""

    def __init__(self, text: str = "", asis: bool = False, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.text = text
        self.asis = asis

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{textwrap.shorten(self.text, 60)}")'


class Error(Text):
    """Error node.

    Notes
    -----
    When the parser encounters an error, this node is created at the location where the
    error is found.  Note this inherits from :class:`Text`; the text contents of this
    Node should indicate where in the source file the error occurred.

    """


class Span(NodeWithChildren):
    newmetakeys: ClassVar[set] = {"strong", "emphas", "little", "insert", "delete"}
    attr_to_tag: ClassVar[dict] = {
        "strong": "strong",
        "emphas": "em",
        "little": "small",
        "insert": "ins",
        "delete": "del",
    }

    def __init__(
        self,
        strong: bool = False,
        emphas: bool = False,
        little: bool = False,
        insert: bool = False,
        delete: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.strong = strong
        self.emphas = emphas
        self.little = little
        self.insert = insert
        self.delete = delete


class Heading(NodeWithChildren):
    newmetakeys: ClassVar[set] = {"title"}

    def __init__(self, title: str = "", **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.title = title


class Manuscript(Heading):
    newmetakeys: ClassVar[set] = {"date"}
    nonum = True

    def __init__(
        self, src: str = "", date: datetime | None = None, **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        self.src = src
        self.date = date

    @property
    def full_number(self) -> str:
        return ""


class Author(Node):
    """An author of the manuscript.

    Notes
    -----
    A Manuscript may have more than one Author node.

    Examples
    --------
    .. rsm::

       :author: {
         :name: Melvin J. Noir
         :affiliation: ACME University
         :email: mel@acme.edu
       }
       ::

    """

    newmetakeys: ClassVar[set] = {
        "name",
        "affiliation",
        "email",
        "orcid",
        "author_note",
    }

    def __init__(
        self,
        name: str = "",
        affiliation: str = "",
        email: str = "",
        orcid: str = "",
        author_note: str = "",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.name: str = name
        """Full name of the author."""
        self.affiliation: str = affiliation
        """Institutional affiliation."""
        self.email: str = email
        """Contact information."""
        self.orcid: str = orcid
        """ORCID identifier."""
        self.author_note: str = author_note
        """Author note (e.g., 'Equal contribution')."""

        self.affiliation_number: int | None = None
        """Affiliation number assigned by transformer (auto-deduped)."""
        self.note_symbol: str = ""
        """Note symbol assigned by transformer (*, †, ‡, §, ¶, ‖, **, ††, etc.)."""


class Abstract(NodeWithChildren):
    """Manuscript abstract.

    Notes
    -----
    By convention, abstracts contain only paragraphs and not other blocks.  This
    convention is not enforced, but it may be assumed to be the case during the
    translation step.

    Examples
    --------
    .. rsm::

       :abstract: {
         :keywords: {cosmology, general relativity, black holes}
       }

       Black holes emit radiation.

       ::

    """

    newmetakeys: ClassVar[set] = {"keywords", "msc"}

    def __init__(
        self,
        keywords: list[str] | None = None,
        msc: list[str] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.keywords = keywords or []
        """Manuscript keywords."""
        self.msc = msc or []
        """Mathematics Subject Classification (MSC) codes."""


class Section(Heading):
    autonumber = True
    level: ClassVar[int] = 2


class Subsection(Section):
    _number_within = Section
    level: ClassVar[int] = 3
    classreftext: ClassVar[str] = "Section {number}"


class Subsubsection(Section):
    _number_within = Subsection
    level: ClassVar[int] = 4
    classreftext: ClassVar[str] = "Section {number}"


class BaseParagraph(Heading):
    """Foo."""


class Paragraph(BaseParagraph):
    pass


class Note(BaseParagraph):
    pass


class Enumerate(NodeWithChildren):
    pass


class Itemize(NodeWithChildren):
    pass


class Keyword(Span):
    pass


class Construct(NodeWithChildren):
    kind_to_keyword: dict[str, str] = {
        "let": "LET",
        "define": "DEFINE",
        "write": "WRITE",
        "case": "CASE",
        "then": "THEN",
        "new": "NEW",
        "assume": "ASSUME",
        "prove": "PROVE",
        "pick": "PICK",
        "suffices": "SUFFICES",
        "claim": "⊢",
        "claimblock": "⊢",
        "qed": "QED",
        "st": "SUCH THAT",
    }

    def __init__(self, kind: str = "", **kwargs: Any):
        super().__init__(**kwargs)
        self.kind = kind

    @property
    def keyword(self):
        return self.kind_to_keyword[self.kind]


class ClaimBlock(Construct):
    def __init__(self, **kwargs: Any):
        super().__init__(kind="claimblock", **kwargs)


class Math(NodeWithChildren):
    pass


class MathBlock(NodeWithChildren):
    autonumber = True
    _number_within = Section
    classreftext: ClassVar[str] = "({number})"
    newmetakeys: ClassVar[set] = {"isclaim"}

    def __init__(self, isclaim: bool = False, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.isclaim = isclaim

    @property
    def long_reftext(self) -> str:
        return f"Equation {self.reftext}".strip()


class SourceCode(Text):
    def __init__(self, text: str = "", lang: str = "", **kwargs: Any) -> None:
        kwargs.pop("asis", None)
        super().__init__(text, asis=True, **kwargs)
        self.lang = lang

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{textwrap.shorten(self.text, 60)}", lang="{self.lang}")'


class Code(NodeWithChildren):
    newmetakeys: ClassVar[set] = {"lang"}

    def __init__(self, source: str = "", lang: str = "", **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.lang = lang
        self.source = SourceCode(source, lang)


class CodeBlock(NodeWithChildren):
    classreftext: ClassVar[str] = "Code Block {number}"
    newmetakeys: ClassVar[set] = {"lang"}

    def __init__(self, source: str = "", lang: str = "", **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.lang = lang
        self.source = SourceCode(source, lang)


class Algorithm(NodeWithChildren):
    r"""Algorithm pseudocode.

    Notes
    -----
    The contents of an Algorithm node are not RSM markup.  They must be written in
    LaTeX-style notation using the `algorithmic` [1]_ environment.


    RSM uses `pseudocode.js` [2]_ to render algorithms in the frontend.

    References
    ----------
    .. [1] https://www.overleaf.com/learn/latex/Algorithms#The_algpseudocode_and_algorithm_packages
    .. [2] https://saswat.padhi.me/pseudocode.js/

    Examples
    --------
    .. rsm::

       :algorithm:
         \begin{algorithm}
         \caption{Quicksort}
         \begin{algorithmic}
         \PROCEDURE{Quicksort}{$A, p, r$}
             \IF{$p < r$}
                 \STATE $q = $ \CALL{Partition}{$A, p, r$}
                 \STATE \CALL{Quicksort}{$A, p, q - 1$}
                 \STATE \CALL{Quicksort}{$A, q + 1, r$}
             \ENDIF
         \ENDPROCEDURE
         \PROCEDURE{Partition}{$A, p, r$}
             \STATE $x = A[r]$
             \STATE $i = p - 1$
             \FOR{$j = p$ \TO $r - 1$}
                 \IF{$A[j] < x$}
                     \STATE $i = i + 1$
                     \STATE exchange
                     $A[i]$ with $A[j]$
                 \ENDIF
                 \STATE exchange $A[i]$ with $A[r]$
             \ENDFOR
         \ENDPROCEDURE
         \end{algorithmic}
         \end{algorithm}
       ::
    """

    autonumber = True


class Appendix(Node):
    """Mark the start of the Appendix sections.

    The Appendix node currently has no visible output in the manuscript.  Instead, it
    affects how subsequent sections are numbered.

    Notes
    -----
    The Appendix must not contain children.  In fact, its parent subclass is
    :class:`Node`, not :class:`NodeWithChildren`.  Sections "in" the Appendix should
    simply appear after an Appendix node.

    Examples
    --------
    In RSM markup, the Appendix is a stamp, i.e. it does not need a closing Halmos.

    .. code-block:: text

       # Before Appendix

       # Before Appendix

       ## Subsec

       :appendix:

       # After Appendix

    The above source is parsed into a Manuscript tree equivalent to the following.

    >>> msc = nodes.Manuscript().append(
    ...     [
    ...         nodes.Section(title="Before Appendix"),
    ...         nodes.Section(title="Before Appendix").append(nodes.Subsection(title="Subsec")),
    ...         nodes.Appendix(),
    ...         nodes.Section(title="After Appendix"),
    ...     ]
    ... )
    >>> print(msc.sexp())
    (Manuscript
      (Section)
      (Section
        (Subsection))
      (Appendix)
      (Section))

    Run the transform step on this Manuscript so the Sections will be autonumbered.

    >>> tform = rsm.transformer.Transformer()
    >>> tform.transform(msc)
    Manuscript(parent=None, [Section, Section, Appendix, Section])

    By default, Sections appearing after the Appendix receive letter numerals.

    >>> for sec in msc.traverse(nodeclass=nodes.Section):
    ...     print(f'{sec.full_number}. {sec.title}')
    1. Before Appendix
    2. Before Appendix
    2.1. Subsec
    A. After Appendix

    """


class BaseReference(Node):
    def __init__(
        self,
        target: str | Node | None = None,
        overwrite_reftext: str = "",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.overwrite_reftext = overwrite_reftext
        self.target = target

    def _attrs_for_repr_and_eq(self) -> list[str]:
        return super()._attrs_for_repr_and_eq() + ["target", "overwrite_reftext"]


class PendingReference(BaseReference):
    def __init__(self, target: str = "", **kwargs: Any) -> None:
        super().__init__(target, **kwargs)

    @property
    def external_file(self) -> str | None:
        """Extract external file path from target, or None if internal reference."""
        if "#" in str(self.target):
            return str(self.target).split("#", 1)[0]
        return None

    @property
    def target_label(self) -> str:
        """Extract label from target."""
        if "#" in str(self.target):
            return str(self.target).split("#", 1)[1]
        return str(self.target)


class Reference(BaseReference):
    def __init__(
        self,
        target: Node | None = None,
        external_file: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(target, **kwargs)
        self.external_file = external_file


class PendingPrev(BaseReference):
    def __init__(self, target: str = "", **kwargs: Any) -> None:
        super().__init__(target, **kwargs)


class URL(BaseReference):
    def __init__(self, target: str = "", **kwargs: Any) -> None:
        super().__init__(target, **kwargs)


class PendingCite(Node):
    def __init__(self, targetlabels: list[str] | None = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.targetlabels = targetlabels or []


class Cite(Node):
    def __init__(self, targets: list[Node] | None = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.targets = targets or []


class Statement(NodeWithChildren):
    pass


class Proof(NodeWithChildren):
    has_handrail = True


class Subproof(NodeWithChildren):  # importantly, NOT a subclass of Proof!
    pass


class Sketch(NodeWithChildren):
    has_handrail = True


class Step(Paragraph):
    autonumber = True
    classreftext: ClassVar[str] = "Step ⟨{number}⟩"
    possible_parents: ClassVar[set[type["NodeWithChildren"]]] = {Proof, Subproof}
    has_handrail = True


Step.possible_parents.add(Step)
Step._number_within = (Proof, Step)


class Theorem(Heading):
    autonumber = True
    title = ""
    _number_within = Section
    newmetakeys: ClassVar[set] = {"title", "goals", "stars", "clocks"}
    has_handrail = True

    def __init__(
        self,
        title: str = "",
        goals: list[BaseReference] | None = None,
        stars: int = 0,
        clocks: int = 0,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.title = title
        self.goals = goals or []
        self.stars = stars
        self.clocks = clocks


class Lemma(Theorem):
    _number_as = Theorem


class Corollary(Theorem):
    _number_as = Theorem


class Example(Theorem):
    _number_as = Theorem


class Exercise(Theorem):
    _number_as = Theorem


class Proposition(Theorem):
    _number_as = Theorem


class Problem(Theorem):
    _number_as = Theorem


class Porism(Theorem):
    _number_as = Theorem


class Remark(Theorem):
    _number_as = Theorem


class Definition(Theorem):
    _number_as = Theorem


class Bibliography(NodeWithChildren):
    pass


class Bibitem(Node):
    autonumber = True
    classreftext: ClassVar[str] = "{number}"
    newmetakeys: ClassVar[set] = {
        "kind",
        "author",
        "title",
        "year",
        "journal",
        "volume",
        "number",
        "publisher",
        "doi",
    }

    def __init__(
        self,
        kind: str = "",
        author: str = "",
        title: str = "",
        year: int = -1,
        journal: str = "",
        volume: int = -1,
        number: int = -1,
        publisher: str = "",
        doi: str = "",
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.kind = kind
        self.author = author
        self.title = title
        self.year = year
        self.journal = journal
        self.volume = volume
        self.number = number
        self.publisher = publisher
        self.doi = doi
        self.backlinks = []


class UnknownBibitem(Bibitem):
    def __init__(self, number: str | int = "?", **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.number = number


class Asset(NodeWithChildren):
    autonumber = True
    _number_within = Section
    newmetakeys: ClassVar[set] = {"path", "scale"}

    def __init__(
        self, path: Path | str = "", scale: float = 1.0, **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        self.path = Path(path)
        self.scale = scale


class Figure(Asset):
    classreftext = "Figure {number}"


class Video(Asset):
    classreftext = "Video {number}"


class Html(Asset):
    classreftext = "Html {number}"


class Draft(NodeWithChildren):
    pass


class Table(NodeWithChildren):
    autonumber = True


class TableHead(NodeWithChildren):
    pass


class TableBody(NodeWithChildren):
    pass


class TableRow(NodeWithChildren):
    pass


class TableDatum(NodeWithChildren):
    pass


class Caption(Paragraph):
    possible_parents: ClassVar[set[type["NodeWithChildren"]]] = {
        Figure,
        Video,
        Html,
        Table,
    }


class Contents(Itemize):
    pass


class Item(BaseParagraph):
    possible_parents: ClassVar[set[type["NodeWithChildren"]]] = {
        Itemize,
        Enumerate,
        Contents,
    }


def all_node_subtypes():
    return [
        obj
        for name, obj in globals().items()
        if type(obj) is type and issubclass(obj, Node)
    ]



from conftest import compare_have_want


def test_simple():
    compare_have_want(
        have="""\
        # The Perron non-backtracking eigenvalue after node addition {
          :label: mylbl
          :date: 2022-03-29
        }

        :author: {
          :name: Leo Torres
          :affiliation: Max Planck Institute for Mathematics in the Sciences
          :email: leo@leotrs.com
        }
        ::

        :abstract: {
            :keywords: {spectral graph theory, non-backtracking, interlacing}
            :msc: {05C50, 05C82, 15A18, 15B99}
          }
          first
          second
          third

          fourth
          fifth

        ::

        ## Introduction {
          :title:
          :label: sec-introduction
          :types: {t1, t2}
        }

        Lorem ipsum.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div id="mylbl" class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>The Perron non-backtracking eigenvalue after node addition</h1>

        <p class="manuscript-date">March 29, 2022</p>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Leo Torres</p>

        <p>Max Planck Institute for Mathematics in the Sciences</p>

        <p><a href="mailto:leo@leotrs.com">leo@leotrs.com</a></p>
        </div>
        </div>

        <div class="abstract" data-nodeid="2">

        <h3>Abstract</h3>

        <div class="paragraph" data-nodeid="3">

        <p>first second third</p>

        </div>

        <div class="paragraph" data-nodeid="5">

        <p>fourth fifth</p>

        </div>

        <p class="keywords">Keywords: spectral graph theory, non-backtracking, interlacing</p>

        <p class="msc">MSC: 05C50, 05C82, 15A18, 15B99</p>

        </div>

        <section id="sec-introduction" class="section level-2 t1 t2" data-nodeid="7">

        <h2>1. Introduction</h2>

        <div class="paragraph" data-nodeid="8">

        <p>Lorem ipsum.</p>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )



from conftest import compare_have_want


def test_display_alone():
    compare_have_want(
        have="""\
        Some math
        :mathblock: {
          :label: eqn-plus
          :types: smallequation
        }
          2 + 2 = 4
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Some math </p>
        <div id="eqn-plus" class="mathblock smallequation" data-nodeid="3">
        $$
        2 + 2 = 4
        $$
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_inline_no_meta():
    compare_have_want(
        have="""\
        This paragraph contains inline math :math: 2 + 2 = 4::.
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This paragraph contains inline math <span class="math" data-nodeid="3">\(2 + 2 = 4\)</span>.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_inline_with_meta():
    compare_have_want(
        have="""\
        This paragraph contains inline math :math: {:label: bar, :types: smallequation} 2 + 2
        = 4::.
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This paragraph contains inline math <span id="bar" class="math smallequation" data-nodeid="3">\(2 + 2
        = 4\)</span>.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_math_with_shortcuts():
    compare_have_want(
        have=r"""
        ## My Section

        When $a \ne 0$, there are two solutions to $ax^2 + bx + c = 0$ and they are
        $$x = {-b \pm \sqrt{b^2-4ac} \over 2a}.$$
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. My Section</h2>

        <div class="paragraph" data-nodeid="2">

        <p>When <span class="math" data-nodeid="4">\(a \ne 0\)</span>, there are two solutions to <span class="math" data-nodeid="7">\(ax^2 + bx + c = 0\)</span> and they are </p>
        <div class="mathblock" data-nodeid="10">
        $$
        x = {-b \pm \sqrt{b^2-4ac} \over 2a}.
        $$
        </div>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_math_ref():
    compare_have_want(
        have=r"""
        ## My Section

        This is some inline $2+2=4$ math.  And then some display math.
        :mathblock:
          {:label: eqn-foo}
          2+2=4
        ::
        And now we refer to :ref:eqn-foo::.
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. My Section</h2>

        <div class="paragraph" data-nodeid="2">

        <p>This is some inline <span class="math" data-nodeid="4">\(2+2=4\)</span> math.  And then some display math. </p>
        <div id="eqn-foo" class="mathblock" data-nodeid="7">
        $$
        2+2=4
        $$
        </div>
        <p> And now we refer to <a class="reference" href="#eqn-foo">(1.1)</a>.</p>
        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_mathblock_nonum():
    compare_have_want(
        have=r"""
        Some math
        :mathblock:
          {:nonum:}
          2+2=4
        ::
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Some math </p>
        <div class="mathblock" data-nodeid="3">
        $$
        2+2=4
        $$
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_mathblock_nonum_with_shortcut():
    compare_have_want(
        have=r"""
        Some math
        $${:nonum:}
          2+2=4
        $$
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Some math </p>
        <div class="mathblock" data-nodeid="3">
        $$
        2+2=4
        $$
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_unclosed_dollar_sign():
    compare_have_want(
        have=r"""
        There are three dollar signs here $2+2 $= 0$.
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>There are three dollar signs here <span class="math" data-nodeid="3">\(2+2\)</span>= 0 [CST error at (0, 43) - (0, 44)] .</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )



from conftest import compare_have_want_handrails


def test_paragraph_with_mathblock_ending_with_text():
    compare_have_want_handrails(
        have="""
        This is a paragraph with math inside of it
        $$
        2+2=4,
        $$
        and it ends with text.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph hr hr-hidden" tabindex=0 data-nodeid="1">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Paragraph</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">

        <p>This is a paragraph with math inside of it </p>
        <div class="mathblock hr hr-hidden hr-offset" tabindex=0 data-nodeid="3">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Equation (1)</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">
        $$
        2+2=4,
        $$
        </div>

        <div class="hr-info-zone">
        <div class="hr-info"><div class="eqn-number"><p>(1)</p></div></div>
        </div>

        </div>
        <p> and it ends with text.</p>
        </div>

        <div class="hr-info-zone">
        <div class="hr-info"></div>
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_paragraph_ending_with_mathblock():
    compare_have_want_handrails(
        have="""
        This is a paragraph with math inside of it
        $$
        2+2=4.
        $$
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph hr hr-hidden" tabindex=0 data-nodeid="1">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Paragraph</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">

        <p>This is a paragraph with math inside of it </p>
        <div class="mathblock hr hr-hidden hr-offset" tabindex=0 data-nodeid="3">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Equation (1)</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">
        $$
        2+2=4.
        $$
        </div>

        <div class="hr-info-zone">
        <div class="hr-info"><div class="eqn-number"><p>(1)</p></div></div>
        </div>

        </div>

        </div>

        <div class="hr-info-zone">
        <div class="hr-info"></div>
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_paragraph_with_icon_ending_with_mathblock():
    compare_have_want_handrails(
        have=r"""
        :paragraph: {:icon: heart} Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed
        do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam,
        quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
        $$
        \left( \mathbf{B} \mathbf{v} \right)_{i \to j} = \sum_{k \to i \in \vec{E}}
        \mathbf{v}_{k \to i} - \mathbf{v}_{j \to i}.
        $$
        """,
        want=r"""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph hr hr-hidden" tabindex=0 data-nodeid="1">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Paragraph</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">

        <p>Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. </p>
        <div class="mathblock hr hr-hidden hr-offset" tabindex=0 data-nodeid="3">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Equation (1)</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">
        $$
        \left( \mathbf{B} \mathbf{v} \right)_{i \to j} = \sum_{k \to i \in \vec{E}}
        \mathbf{v}_{k \to i} - \mathbf{v}_{j \to i}.
        $$
        </div>

        <div class="hr-info-zone">
        <div class="hr-info"><div class="eqn-number"><p>(1)</p></div></div>
        </div>

        </div>

        </div>

        <div class="hr-info-zone">
        <div class="hr-info">
                        <div class="icon heart">
                        </div>
                        </div>
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_paragraph_with_icon_with_mathblock_ending_with_text():
    compare_have_want_handrails(
        have=r"""
        :paragraph: {:icon: heart} Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed
        do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam,
        quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
        $$
        \left( \mathbf{B} \mathbf{v} \right)_{i \to j} = \sum_{k \to i \in \vec{E}}
        \mathbf{v}_{k \to i} - \mathbf{v}_{j \to i}.
        $$
        Lorem ipsum.
        """,
        want=r"""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph hr hr-hidden" tabindex=0 data-nodeid="1">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Paragraph</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">

        <p>Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. </p>
        <div class="mathblock hr hr-hidden hr-offset" tabindex=0 data-nodeid="3">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Equation (1)</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">
        $$
        \left( \mathbf{B} \mathbf{v} \right)_{i \to j} = \sum_{k \to i \in \vec{E}}
        \mathbf{v}_{k \to i} - \mathbf{v}_{j \to i}.
        $$
        </div>

        <div class="hr-info-zone">
        <div class="hr-info"><div class="eqn-number"><p>(1)</p></div></div>
        </div>

        </div>
        <p> Lorem ipsum.</p>
        </div>

        <div class="hr-info-zone">
        <div class="hr-info">
                        <div class="icon heart">
                        </div>
                        </div>
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_two_paragraphs_with_icon():
    compare_have_want_handrails(
        have=r"""
        :paragraph: {:icon: heart} Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed
        do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam,
        quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
        $$
        \left( \mathbf{B} \mathbf{v} \right)_{i \to j} = \sum_{k \to i \in \vec{E}}
        \mathbf{v}_{k \to i} - \mathbf{v}_{j \to i}.
        $$
        Lorem ipsum.

        :paragraph: {:icon: success} Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim
        veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo
        consequat.
        """,
        want=r"""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph hr hr-hidden" tabindex=0 data-nodeid="1">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Paragraph</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">

        <p>Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. </p>
        <div class="mathblock hr hr-hidden hr-offset" tabindex=0 data-nodeid="3">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Equation (1)</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">
        $$
        \left( \mathbf{B} \mathbf{v} \right)_{i \to j} = \sum_{k \to i \in \vec{E}}
        \mathbf{v}_{k \to i} - \mathbf{v}_{j \to i}.
        $$
        </div>

        <div class="hr-info-zone">
        <div class="hr-info"><div class="eqn-number"><p>(1)</p></div></div>
        </div>

        </div>
        <p> Lorem ipsum.</p>
        </div>

        <div class="hr-info-zone">
        <div class="hr-info">
                        <div class="icon heart">
                        </div>
                        </div>
        </div>

        </div>

        <div class="paragraph hr hr-hidden" tabindex=0 data-nodeid="6">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Paragraph</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">

        <p>Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.</p>

        </div>

        <div class="hr-info-zone">
        <div class="hr-info">
                        <div class="icon success">
                        </div>
                        </div>
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_two_paragraphs_with_icon_no_mathblock():
    compare_have_want_handrails(
        have=r"""
        :paragraph: {:icon: heart} Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed
        do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam,
        quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.

        :paragraph: {:icon: success} Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim
        veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo
        consequat.
        """,
        want=r"""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph hr hr-hidden" tabindex=0 data-nodeid="1">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Paragraph</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">

        <p>Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.</p>

        </div>

        <div class="hr-info-zone">
        <div class="hr-info">
                        <div class="icon heart">
                        </div>
                        </div>
        </div>

        </div>

        <div class="paragraph hr hr-hidden" tabindex=0 data-nodeid="3">

        <div class="hr-collapse-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-menu-zone">

        <div class="hr-menu">

          <div class="hr-menu-label">
            <span class="hr-menu-item-text">Paragraph</span>
          </div>

          <div class="hr-menu-separator"></div>

          <div class="hr-menu-item link disabled">
            <span class="icon link">
            </span>
            <span class="hr-menu-item-text">Copy link</span>
          </div>

          <div class="hr-menu-item">
            <span class="icon code">
            </span>
            <span class="hr-menu-item-text">Source</span>
          </div>

        </div>

        </div>

        <div class="hr-border-zone">

                        <div class="hr-border-dots">
                          <div class="icon dots">
                          </div>
                        </div>
                        <div class="hr-border-rect">
                        </div>

        </div>

        <div class="hr-spacer-zone">
        <div class="hr-spacer"></div>
        </div>

        <div class="hr-content-zone">

        <p>Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.</p>

        </div>

        <div class="hr-info-zone">
        <div class="hr-info">
                        <div class="icon success">
                        </div>
                        </div>
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )



from conftest import compare_have_want


def test_real_life_example():
    compare_have_want(
        have=r"""
        # A simple proof

        :let: $G$ be a simple graph::.  We will prove a fundamental fact about its structure.
        :define: $L(G) := |E(G)|$, where $E(G)$ is the set of edges of $G$:: and :write:
        $\deg(u)$ for the degree of a node $u \in V(G)$::.

        :theorem: {
          :label: thm-main
          :goal: eqn-thm
          :types: {important, main}
        }

          We have
          $${:label: eqn-thm}
            \sum_{u \in V(G)} \deg(u) = 2L(G).
          $$

        ::

        :proof:

          :step: :define: $n = |V(G)|$.:: ::

          :step: :case: $n = 1$.::

            :step: :claim: $L(G) = 0$ ::.

              :p: $G$ has no self-loops by assumption.::

            ::

            :step: :let: $u \in V(G)$::, :then: $\deg(u) = 0$::.

              :p: Since $G$ has only one node, and it has no self-loops.::

            ::

            :step: :qed:
            ::

          ::

          :step: :let: $v \notin V(G)$::.  :write: $H$ for the graph resulting from adding $v$
            to $G$::. SUFFICES :assume: :span: :span:{:label: asm} the theorem is true for $G$
            ::::::, :prove: the theorem is true for $H$::.

            :p: By induction. ::

          ::

          :step: :define: $\deg_G(u), \deg_H(u)$ as the degree of node $u$ in
            $G$ and $H$ respectively.:: Note $\deg_G(v) = 0$. ::

          :step:{:label: stp-edges}

            :claim: $L(H) = L(G) + \deg_H(v)$::.

            :p: Because the only added edges are those incident to $v$.::

          ::

          :step:{:label: stp-sum}
            :claim: $\sum_{u \in V(H)} \deg_H(u) = 2\deg_H(v) + \sum_{u \in V(G)} \deg_G(u)$::.

            :step: :write: $W \subset V(G)$ for the set of neighbors of $v$.::::

            :step: We have
              $$
                \begin{align}
                \sum_{u \in V(H)} \deg_H(u) &= \phantom{2} \deg_H(v) + \sum_{u \in V(H)\setminus\{v\}} \deg_H(u) \\
                \sum_{u \in V(H)} \deg_H(u) &= \phantom{2} \deg_H(v) + \sum_{u \in W} \deg_H(u) + \sum_{u \in V(H) \setminus W \setminus \{v\}} \deg_H(u) \\
                \sum_{u \in V(H)} \deg_H(u) &= \phantom{2} \deg_H(v) + \sum_{u \in W} (\deg_G(u) + 1) + \sum_{u \in V(H) \setminus W \setminus \{v\}} \deg_G(u) \\
                \sum_{u \in V(H)} \deg_H(u) &= 2\deg_H(v) + \sum_{u \in W} \deg_G(u) + \sum_{u \in V(H) \setminus W \setminus \{v\}} \deg_G(u) \\
                \sum_{u \in V(H)} \deg_H(u) &= 2\deg_H(v) + \sum_{u \in V(G)} \deg_G(u),
                \end{align}
              $$

              where in the right-most term in third equation we have used the fact that for
              every node $u$ in $V(G) \setminus W \setminus v$ we have $\deg_H(u) = \deg_G(u)$.

            ::

          ::

          :step: :qed:

            :p: We have
              $$
                \begin{align}
                2 L(G) &= \sum_{u \in V(G)} \deg_G(u) \\
                2 L(G) + 2 \deg_H(v) &=  2 \deg_H(v) + \sum_{u \in V(G)} \deg_G(u) \\
                2 L(H) &=  \sum_{u \in V(H)} \deg_H(u),
                \end{align}
              $$

              where the first equation is true by :ref:asm,induction hypothesis::, and the last
              equation is due to :prev2: and :prev:.

            ::

          ::
        ::
        """,
        want=r"""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>A simple proof</h1>

        <div class="paragraph" data-nodeid="1">

        <p><span class="construct let assumption" data-nodeid="2"><span class="keyword" data-nodeid="3">LET</span> <span class="math" data-nodeid="6">\(G\)</span> be a simple graph</span>.  We will prove a fundamental fact about its structure. <span class="construct define assumption" data-nodeid="10"><span class="keyword" data-nodeid="11">DEFINE</span> <span class="math" data-nodeid="14">\(L(G) := |E(G)|\)</span>, where <span class="math" data-nodeid="17">\(E(G)\)</span> is the set of edges of <span class="math" data-nodeid="20">\(G\)</span></span> and <span class="construct write assumption" data-nodeid="23"><span class="keyword" data-nodeid="24">WRITE</span> <span class="math" data-nodeid="27">\(\deg(u)\)</span> for the degree of a node <span class="math" data-nodeid="30">\(u \in V(G)\)</span></span>.</p>

        </div>

        <div id="thm-main" class="theorem important main" data-nodeid="33">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.</span></p>

        </div>

        <div class="paragraph" data-nodeid="34">

        <p>We have </p>
        <div id="eqn-thm" class="mathblock" data-nodeid="36">
        $$
        \sum_{u \in V(G)} \deg(u) = 2L(G).
        $$
        </div>

        </div>

        </div>

        <div class="proof" data-nodeid="38">

        <div class="paragraph hr-label">

        <p><span class="span label">Proof.</span></p>

        </div>

        <div class="step" data-nodeid="39">

        <div class="statement" data-nodeid="40">

        <div class="paragraph" data-nodeid="41">

        <p><span class="construct define assumption" data-nodeid="42"><span class="keyword" data-nodeid="43">DEFINE</span> <span class="math" data-nodeid="46">\(n = |V(G)|\)</span>.</span></p>

        </div>

        </div>

        </div>

        <div class="step" data-nodeid="49">

        <div class="statement" data-nodeid="50">

        <div class="paragraph" data-nodeid="51">

        <p><span class="construct case assumption" data-nodeid="52"><span class="keyword" data-nodeid="53">CASE</span> <span class="math" data-nodeid="56">\(n = 1\)</span>.</span></p>

        </div>

        </div>

        <div class="subproof" data-nodeid="59">

        <div class="step" data-nodeid="60">

        <div class="statement" data-nodeid="61">

        <div class="paragraph" data-nodeid="62">

        <p><span class="construct claim" data-nodeid="63"><span class="keyword" data-nodeid="64">⊢</span> <span class="math" data-nodeid="67">\(L(G) = 0\)</span></span>.</p>

        </div>

        </div>

        <div class="subproof" data-nodeid="70">

        <div class="paragraph" data-nodeid="71">

        <p><span class="math" data-nodeid="72">\(G\)</span> has no self-loops by assumption.</p>

        </div>

        </div>

        </div>

        <div class="step" data-nodeid="75">

        <div class="statement" data-nodeid="76">

        <div class="paragraph" data-nodeid="77">

        <p><span class="construct let assumption" data-nodeid="78"><span class="keyword" data-nodeid="79">LET</span> <span class="math" data-nodeid="82">\(u \in V(G)\)</span></span>, <span class="construct then" data-nodeid="85"><span class="keyword" data-nodeid="86">THEN</span> <span class="math" data-nodeid="89">\(\deg(u) = 0\)</span></span>.</p>

        </div>

        </div>

        <div class="subproof" data-nodeid="92">

        <div class="paragraph" data-nodeid="93">

        <p>Since <span class="math" data-nodeid="95">\(G\)</span> has only one node, and it has no self-loops.</p>

        </div>

        </div>

        </div>

        <div class="step" data-nodeid="98">

        <div class="statement" data-nodeid="99">

        <div class="paragraph" data-nodeid="100">

        <p><span class="construct qed" data-nodeid="101"><span class="keyword" data-nodeid="102">QED</span></span></p>

        </div>

        </div>

        </div>

        </div>

        </div>

        <div class="step" data-nodeid="104">

        <div class="statement" data-nodeid="105">

        <div class="paragraph" data-nodeid="106">

        <p><span class="construct let assumption" data-nodeid="107"><span class="keyword" data-nodeid="108">LET</span> <span class="math" data-nodeid="111">\(v \notin V(G)\)</span></span>. <span class="construct write assumption" data-nodeid="114"><span class="keyword" data-nodeid="115">WRITE</span> <span class="math" data-nodeid="118">\(H\)</span> for the graph resulting from adding <span class="math" data-nodeid="121">\(v\)</span> to <span class="math" data-nodeid="124">\(G\)</span></span>. SUFFICES <span class="construct assume assumption" data-nodeid="127"><span class="keyword" data-nodeid="128">ASSUME</span><span class="span" data-nodeid="130"><span id="asm" class="span" data-nodeid="131">the theorem is true for <span class="math" data-nodeid="133">\(G\)</span></span></span></span>, <span class="construct prove" data-nodeid="136"><span class="keyword" data-nodeid="137">PROVE</span> the theorem is true for <span class="math" data-nodeid="140">\(H\)</span></span>.</p>

        </div>

        </div>

        <div class="subproof" data-nodeid="143">

        <div class="paragraph" data-nodeid="144">

        <p>By induction.</p>

        </div>

        </div>

        </div>

        <div class="step" data-nodeid="146">

        <div class="statement" data-nodeid="147">

        <div class="paragraph" data-nodeid="148">

        <p><span class="construct define assumption" data-nodeid="149"><span class="keyword" data-nodeid="150">DEFINE</span> <span class="math" data-nodeid="153">\(\deg_G(u), \deg_H(u)\)</span> as the degree of node <span class="math" data-nodeid="156">\(u\)</span> in <span class="math" data-nodeid="159">\(G\)</span> and <span class="math" data-nodeid="162">\(H\)</span> respectively.</span> Note <span class="math" data-nodeid="166">\(\deg_G(v) = 0\)</span>.</p>

        </div>

        </div>

        </div>

        <div id="stp-edges" class="step" data-nodeid="169">

        <div class="statement" data-nodeid="170">

        <div class="paragraph" data-nodeid="171">

        <p><span class="construct claim" data-nodeid="172"><span class="keyword" data-nodeid="173">⊢</span> <span class="math" data-nodeid="176">\(L(H) = L(G) + \deg_H(v)\)</span></span>.</p>

        </div>

        </div>

        <div class="subproof" data-nodeid="179">

        <div class="paragraph" data-nodeid="180">

        <p>Because the only added edges are those incident to <span class="math" data-nodeid="182">\(v\)</span>.</p>

        </div>

        </div>

        </div>

        <div id="stp-sum" class="step" data-nodeid="185">

        <div class="statement" data-nodeid="186">

        <div class="paragraph" data-nodeid="187">

        <p><span class="construct claim" data-nodeid="188"><span class="keyword" data-nodeid="189">⊢</span> <span class="math" data-nodeid="192">\(\sum_{u \in V(H)} \deg_H(u) = 2\deg_H(v) + \sum_{u \in V(G)} \deg_G(u)\)</span></span>.</p>

        </div>

        </div>

        <div class="subproof" data-nodeid="195">

        <div class="step" data-nodeid="196">

        <div class="statement" data-nodeid="197">

        <div class="paragraph" data-nodeid="198">

        <p><span class="construct write assumption" data-nodeid="199"><span class="keyword" data-nodeid="200">WRITE</span> <span class="math" data-nodeid="203">\(W \subset V(G)\)</span> for the set of neighbors of <span class="math" data-nodeid="206">\(v\)</span>.</span></p>

        </div>

        </div>

        </div>

        <div class="step" data-nodeid="209">

        <div class="statement" data-nodeid="210">

        <div class="paragraph" data-nodeid="211">

        <p>We have </p>
        <div class="mathblock" data-nodeid="213">
        $$
        \begin{align}
                \sum_{u \in V(H)} \deg_H(u) &= \phantom{2} \deg_H(v) + \sum_{u \in V(H)\setminus\{v\}} \deg_H(u) \\
                \sum_{u \in V(H)} \deg_H(u) &= \phantom{2} \deg_H(v) + \sum_{u \in W} \deg_H(u) + \sum_{u \in V(H) \setminus W \setminus \{v\}} \deg_H(u) \\
                \sum_{u \in V(H)} \deg_H(u) &= \phantom{2} \deg_H(v) + \sum_{u \in W} (\deg_G(u) + 1) + \sum_{u \in V(H) \setminus W \setminus \{v\}} \deg_G(u) \\
                \sum_{u \in V(H)} \deg_H(u) &= 2\deg_H(v) + \sum_{u \in W} \deg_G(u) + \sum_{u \in V(H) \setminus W \setminus \{v\}} \deg_G(u) \\
                \sum_{u \in V(H)} \deg_H(u) &= 2\deg_H(v) + \sum_{u \in V(G)} \deg_G(u),
                \end{align}
        $$
        </div>

        </div>

        <div class="paragraph" data-nodeid="215">

        <p>where in the right-most term in third equation we have used the fact that for every node <span class="math" data-nodeid="217">\(u\)</span> in <span class="math" data-nodeid="220">\(V(G) \setminus W \setminus v\)</span> we have <span class="math" data-nodeid="223">\(\deg_H(u) = \deg_G(u)\)</span>.</p>

        </div>

        </div>

        </div>

        </div>

        </div>

        <div class="step last" data-nodeid="226">

        <div class="statement" data-nodeid="227">

        <div class="paragraph" data-nodeid="228">

        <p><span class="construct qed" data-nodeid="229"><span class="keyword" data-nodeid="230">QED</span></span></p>

        </div>

        </div>

        <div class="subproof" data-nodeid="232">

        <div class="paragraph" data-nodeid="233">

        <p>We have </p>
        <div class="mathblock" data-nodeid="235">
        $$
        \begin{align}
                2 L(G) &= \sum_{u \in V(G)} \deg_G(u) \\
                2 L(G) + 2 \deg_H(v) &=  2 \deg_H(v) + \sum_{u \in V(G)} \deg_G(u) \\
                2 L(H) &=  \sum_{u \in V(H)} \deg_H(u),
                \end{align}
        $$
        </div>

        </div>

        <div class="paragraph" data-nodeid="237">

        <p>where the first equation is true by <a class="reference" href="#asm">induction hypothesis</a>, and the last equation is due to <a class="reference" href="#stp-edges">Step ⟨5⟩</a> and <a class="reference" href="#stp-sum">Step ⟨6⟩</a>.</p>

        </div>

        </div>

        </div>

        <div class="halmos"></div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_construct_at_line_start():
    """Regression test: construct tags at start of line should start new construct, not be consumed as TEXT."""
    compare_have_want(
        have=r"""
        :let: $G$ be a simple graph::.  We will prove a fact.
        :define: $L(G) := |E(G)|$, where $E(G)$ is the set of edges::.

        More text here.
        """,
        want=r"""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="construct let assumption" data-nodeid="2"><span class="keyword" data-nodeid="3">LET</span> <span class="math" data-nodeid="6">\(G\)</span> be a simple graph</span>.  We will prove a fact. <span class="construct define assumption" data-nodeid="10"><span class="keyword" data-nodeid="11">DEFINE</span> <span class="math" data-nodeid="14">\(L(G) := |E(G)|\)</span>, where <span class="math" data-nodeid="17">\(E(G)\)</span> is the set of edges</span>.</p>

        </div>

        <div class="paragraph" data-nodeid="21">

        <p>More text here.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )



import subprocess
import sys
from importlib.metadata import version
from textwrap import dedent

import pytest


def cmd(src):
    return " ".join(["rsm", "render", f"'{src}'"])


@pytest.mark.slow
def test_render(tmp_path):
    want = dedent(
        """
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Foo.</p>

        </div>

        <div class="paragraph" data-nodeid="3">

        <p>Bar.</p>

        </div>

        <div class="paragraph" data-nodeid="5">

        <p>Baz.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """
    )
    # In this source code, we use the \r\n line ending typical in Windows systems.  When
    # running this test on a UNIX machine, they will simply be ignored.  When running on
    # a Windows machine, the test will fail if they are not present.
    src = "Foo.\r\n\r\nBar.\r\n\r\nBaz.\r\n"
    file = tmp_path / "test.rsm"
    file.write_text(src)
    result = subprocess.run(
        f"rsm render {str(file.resolve())}",
        stdout=subprocess.PIPE,
        check=True,
        shell=True,
    )

    # get rid of any logs appearing before the html body, and decode
    output = result.stdout[result.stdout.find(b"<body>") :].decode("utf-8")
    if sys.platform == "win32":
        output = output.replace("\r", "")
    assert output.strip() == want.strip()


@pytest.mark.slow
def test_invalid_rsm():
    have = "test_file.rsm"
    with pytest.raises(subprocess.CalledProcessError):
        subprocess.run(cmd(have), stdout=subprocess.PIPE, check=True, shell=True)


@pytest.mark.slow
def test_render_version():
    result = subprocess.run(
        "rsm --version", stdout=subprocess.PIPE, check=True, shell=True
    )
    output = result.stdout.decode("utf-8").strip()
    assert output == f"rsm-markup v{version('rsm-markup')}"



import pytest

import rsm
from conftest import compare_have_want


def test_simple():
    compare_have_want(
        have="""\
        # My Title

        :enumerate:
          {:label: enm-foo}

          :-: {:label: itm-1} Foo bar.

          :-: {:label: itm-2} Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet Lorem
            ipsum dolor sit amet.

          :-: {:label: itm-3}
          Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet Lorem
          ipsum dolor sit amet.

        ::

        :itemize:
          {:label: itm-foo}

          :-: {:label: itm-4} Foo bar.

          :-: {:label: itm-5} Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet Lorem
            ipsum dolor sit amet.

          :-: {:label: itm-6}
          Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet Lorem
          ipsum dolor sit amet.

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <ol id="enm-foo" class="enumerate" data-nodeid="1">

        <li id="itm-1" class="item" data-nodeid="2">
        Foo bar.
        </li>

        <li id="itm-2" class="item" data-nodeid="4">
        Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet.
        </li>

        <li id="itm-3" class="item" data-nodeid="6">
        Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet.
        </li>

        </ol>

        <ul id="itm-foo" class="itemize" data-nodeid="8">

        <li id="itm-4" class="item" data-nodeid="9">
        Foo bar.
        </li>

        <li id="itm-5" class="item" data-nodeid="11">
        Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet.
        </li>

        <li id="itm-6" class="item" data-nodeid="13">
        Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet Lorem ipsum dolor sit amet.
        </li>

        </ul>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_item_with_wrong_parent():
    with pytest.raises(rsm.nodes.RSMNodeError):
        compare_have_want(
            have="""\
            :-: Foo bar.
            """,
            want="XXX",
        )

    with pytest.raises(rsm.nodes.RSMNodeError):
        compare_have_want(
            have="""\
            # Some section

            :-: Foo bar.
            """,
            want="XXX",
        )

    compare_have_want(
        have="""\
        ## Some section

        Lorem ipsum - Foo bar.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. Some section</h2>

        <div class="paragraph" data-nodeid="2">

        <p>Lorem ipsum - Foo bar.</p>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )



"""Input: Path to .rsm file -- Output: file contents.

Locate a file in disk and read its contents.

"""

import logging
from pathlib import Path

logger = logging.getLogger("RSM").getChild("read ")


class Reader:
    """Read the contents of a RSM source file."""

    def __init__(self) -> None:
        self.path: Path = Path()
        """The path of the RSM source file."""

        self.src: str = ""
        """The contents of the RSM source file."""

    def read(self, path: Path | None) -> str:
        """Read the contents of the path.

        Parameters
        ----------
        path
            Location of the RSM file.

        Returns
        -------
        src
            Contents of the file.

        Notes
        -----
        *path* is stored as :attr:`~Reader.path` and *src* is stored as :attr:`~Reader.src`.

        """
        logger.info("Reading...")
        if not path:
            raise TypeError("Expected path, got None")
        self.path = Path(path)
        with open(path) as file:
            self.src = file.read()
        return self.src



import re
from pathlib import Path

import rsm.tags

ROOT_DIR = Path(__file__).parent.parent
GRAMMAR_FILE = ROOT_DIR / "tree-sitter-rsm/grammar.js"
TAGS_FILE = ROOT_DIR / "rsm/tags.py"


def extract_grammar_tags():
    with open(GRAMMAR_FILE) as f:
        contents = f.read()
    tags = set(re.findall(r":([\w-]+):", contents))
    tags = {tag.replace("-", "_") for tag in tags}
    # :-: is the list item marker, not a traditional tag - the node type is 'item'
    if "_" in tags:
        tags.remove("_")
        tags.add("item")
    return tags


def extract_documented_tags():
    tags = {
        t
        for t in dir(rsm.tags)
        if isinstance(
            getattr(rsm.tags, t),
            rsm.tags.BlockTagInfo
            | rsm.tags.InlineTagInfo
            | rsm.tags.MathTagInfo
            | rsm.tags.MetaTagInfo
            | rsm.tags.ParagraphTagInfo
            | rsm.tags.TableTagInfo,
        )
    }
    return tags


def test_all_tags_are_documented():
    grammar_tags = extract_grammar_tags()
    documented_tags = extract_documented_tags()
    assert grammar_tags == documented_tags



"""Integration tests for external cross-file references."""

from pathlib import Path

import rsm


def test_external_ref_html_generation():
    """External references should generate correct HTML with external class."""
    root_dir = Path(__file__).parent / "fixtures/crossref"

    src = """# Main Document

This document references :ref:definitions/def.rsm#test-def::.
"""

    # Parse
    parser = rsm.tsparser.TSParser()
    tree = parser.parse(src)

    # Transform with root_dir
    from rsm.transformer import Transformer

    t = Transformer(root_dir=root_dir)
    transformed = t.transform(tree)

    # Translate to HTML
    from rsm.translator import Translator

    translator = Translator()
    html = translator.translate(transformed)

    # Verify the HTML contains the correct href and external class
    assert 'href="definitions/def.html#test-def"' in html
    assert 'class="reference external"' in html or 'class="external reference"' in html


def test_internal_ref_html_generation():
    """Internal references should generate normal HTML without external class."""
    src = """# Document

## Section
  {:label: my-section}

This references :ref:my-section::.
"""

    # Parse
    parser = rsm.tsparser.TSParser()
    tree = parser.parse(src)

    # Transform
    from rsm.transformer import Transformer

    t = Transformer()
    transformed = t.transform(tree)

    # Translate to HTML
    from rsm.translator import Translator

    translator = Translator()
    html = translator.translate(transformed)

    # Verify internal ref doesn't have external class
    assert 'href="#my-section"' in html
    assert (
        "external" not in html or '<span class="external"' in html
    )  # Allow word "external" in other contexts


def test_external_ref_missing_file():
    """External references to missing files should generate error nodes."""
    root_dir = Path(__file__).parent / "fixtures/crossref"

    src = """# Document

This references :ref:nonexistent/file.rsm#label::.
"""

    # Parse
    parser = rsm.tsparser.TSParser()
    tree = parser.parse(src)

    # Transform with root_dir
    from rsm.transformer import Transformer

    t = Transformer(root_dir=root_dir)
    transformed = t.transform(tree)

    # Should have an Error node
    errors = [
        node for node in transformed.traverse() if node.__class__.__name__ == "Error"
    ]
    assert len(errors) == 1
    # Error message should mention the missing file
    if errors[0].children:
        assert "nonexistent/file.rsm" in str(errors[0].children[0])
    # If no children, just verify an error was created
    assert errors[0].__class__.__name__ == "Error"



from conftest import compare_have_want


def test_period_after():
    compare_have_want(
        have="""
        Period after math :math:2+2=4::.

        Period after code :code:k=v::.

        Period after claim :claim:foo::.
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Period after math <span class="math" data-nodeid="3">\(2+2=4\)</span>.</p>

        </div>

        <div class="paragraph" data-nodeid="6">

        <p>Period after code <span class="code" data-nodeid="8"><code>k=v</code></span>.</p>

        </div>

        <div class="paragraph" data-nodeid="11">

        <p>Period after claim <span class="construct claim" data-nodeid="13"><span class="keyword" data-nodeid="14">⊢</span> foo</span>.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_period_after_with_shortcut():
    compare_have_want(
        have="""
        Period after math $2+2=4$.

        Period after code `k=v`.

        Period after claim :|-:foo::.
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Period after math <span class="math" data-nodeid="3">\(2+2=4\)</span>.</p>

        </div>

        <div class="paragraph" data-nodeid="6">

        <p>Period after code <span class="code" data-nodeid="8"><code>k=v</code></span>.</p>

        </div>

        <div class="paragraph" data-nodeid="11">

        <p>Period after claim <span class="construct claim" data-nodeid="13"><span class="keyword" data-nodeid="14">⊢</span> foo</span>.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_separate_words():
    compare_have_want(
        have="""
        :span:{:strong:}Separate:: words.

        Separate :span:{:strong:}words::.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="span" data-nodeid="2"><strong>Separate</strong></span> words.</p>

        </div>

        <div class="paragraph" data-nodeid="5">

        <p>Separate <span class="span" data-nodeid="7"><strong>words</strong></span>.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_separate_words_with_shortcut():
    compare_have_want(
        have="""
        **Separate** words.

        Separate **words**.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="span" data-nodeid="2"><strong>Separate</strong></span> words.</p>

        </div>

        <div class="paragraph" data-nodeid="5">

        <p>Separate <span class="span" data-nodeid="7"><strong>words</strong></span>.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_span_multiline_content_middle_of_line():
    compare_have_want(
        have="""
        This is a paragraph
        with :span: {:strong:} a
        span :: that takes multiple lines
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is a paragraph with <span class="span" data-nodeid="3"><strong>a span</strong></span> that takes multiple lines</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_span_multiline_content_beginning_of_line():
    compare_have_want(
        have="""
        This is a paragraph with
        :span: {:strong:} a
        span :: that takes multiple lines
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is a paragraph with <span class="span" data-nodeid="3"><strong>a span</strong></span> that takes multiple lines</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_span_multiline_content_beginning_of_line_with_shortcut():
    compare_have_want(
        have="""
        This is a paragraph with
        ** a
        span ** that takes multiple lines
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is a paragraph with <span class="span" data-nodeid="3"><strong>a span</strong></span> that takes multiple lines</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_span_multiline_content_middle_of_line_with_shortcut():
    compare_have_want(
        have="""
        This is a paragraph
        with ** a
        span ** that takes multiple lines
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is a paragraph with <span class="span" data-nodeid="3"><strong>a span</strong></span> that takes multiple lines</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_math_start_of_line():
    compare_have_want(
        have="""
        This is a paragraph with
        :math:2+2=4:: at the start of a line.
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is a paragraph with <span class="math" data-nodeid="3">\(2+2=4\)</span> at the start of a line.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_math_start_of_line_with_shortcut():
    compare_have_want(
        have="""
        This is a paragraph with
        $2+2=4$ at the start of a line.
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is a paragraph with <span class="math" data-nodeid="3">\(2+2=4\)</span> at the start of a line.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_table_tr():
    compare_have_want(
        have="""
        :table:

        :tbody:

        :tr: :td: *foo* :: :td: **bar** :: ::

        ::

        ::
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="table-wrapper">

        <table class="table" data-nodeid="1">

        <tbody class="tablebody" data-nodeid="2">

        <tr class="tablerow" data-nodeid="3">

        <td class="tabledatum" data-nodeid="4"><span class="span" data-nodeid="5"><em>foo</em></span></td>

        <td class="tabledatum" data-nodeid="7"><span class="span" data-nodeid="8"><strong>bar</strong></span></td>

        </tr>

        </tbody>

        </table>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_table_trshort():
    compare_have_want(
        have="""
        :table:

        :tbody:

        :tr: *foo* : **bar** ::

        ::

        ::
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="table-wrapper">

        <table class="table" data-nodeid="1">

        <tbody class="tablebody" data-nodeid="2">

        <tr class="tablerow" data-nodeid="3">

        <td class="tabledatum" data-nodeid="4"><span class="span" data-nodeid="5"><em>foo</em></span></td>

        <td class="tabledatum" data-nodeid="7"><span class="span" data-nodeid="8"><strong>bar</strong></span></td>

        </tr>

        </tbody>

        </table>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_ignore_space_within_inline():
    compare_have_want(
        have="""
        Ignore space :span: {:strong:} within :: inline.

        Ignore space :span: {:strong:}   within   :: inline.

        Ignore space :claim: within :: inline.

        Ignore space :claim:   within   :: inline.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Ignore space <span class="span" data-nodeid="3"><strong>within</strong></span> inline.</p>

        </div>

        <div class="paragraph" data-nodeid="6">

        <p>Ignore space <span class="span" data-nodeid="8"><strong>within</strong></span> inline.</p>

        </div>

        <div class="paragraph" data-nodeid="11">

        <p>Ignore space <span class="construct claim" data-nodeid="13"><span class="keyword" data-nodeid="14">⊢</span> within</span> inline.</p>

        </div>

        <div class="paragraph" data-nodeid="18">

        <p>Ignore space <span class="construct claim" data-nodeid="20"><span class="keyword" data-nodeid="21">⊢</span> within</span> inline.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_ignore_space_within_inline_with_shortcut():
    compare_have_want(
        have="""
        Ignore space ** within ** inline.

        Ignore space **   within   ** inline.

        Ignore space :|-: within ::. inline.

        Ignore space :|-:   within   ::. inline.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Ignore space <span class="span" data-nodeid="3"><strong>within</strong></span> inline.</p>

        </div>

        <div class="paragraph" data-nodeid="6">

        <p>Ignore space <span class="span" data-nodeid="8"><strong>within</strong></span> inline.</p>

        </div>

        <div class="paragraph" data-nodeid="11">

        <p>Ignore space <span class="construct claim" data-nodeid="13"><span class="keyword" data-nodeid="14">⊢</span> within</span>. inline.</p>

        </div>

        <div class="paragraph" data-nodeid="18">

        <p>Ignore space <span class="construct claim" data-nodeid="20"><span class="keyword" data-nodeid="21">⊢</span> within</span>. inline.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_consecutive_lines_should_be_joined_by_spaces():
    compare_have_want(
        have="""
        Consecutive lines should be
        joined by spaces.

        Consecutive lines should :span:{:strong:}be
        joined:: by spaces.

        Consecutive lines should **be
        joined** by spaces.

        Consecutive lines should be
        :span:{:strong:}joined:: by spaces.

        Consecutive lines should be
        **joined** by spaces.

        Consecutive lines should **be**
        joined by spaces.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Consecutive lines should be joined by spaces.</p>

        </div>

        <div class="paragraph" data-nodeid="3">

        <p>Consecutive lines should <span class="span" data-nodeid="5"><strong>be joined</strong></span> by spaces.</p>

        </div>

        <div class="paragraph" data-nodeid="8">

        <p>Consecutive lines should <span class="span" data-nodeid="10"><strong>be joined</strong></span> by spaces.</p>

        </div>

        <div class="paragraph" data-nodeid="13">

        <p>Consecutive lines should be <span class="span" data-nodeid="15"><strong>joined</strong></span> by spaces.</p>

        </div>

        <div class="paragraph" data-nodeid="18">

        <p>Consecutive lines should be <span class="span" data-nodeid="20"><strong>joined</strong></span> by spaces.</p>

        </div>

        <div class="paragraph" data-nodeid="23">

        <p>Consecutive lines should <span class="span" data-nodeid="25"><strong>be</strong></span> joined by spaces.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_span_span():
    compare_have_want(
        have="""
        :span: foo :: :span: bar ::
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="span" data-nodeid="2">foo</span> <span class="span" data-nodeid="5">bar</span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_span_construct():
    compare_have_want(
        have="""
        :span: foo :: :prove: bar ::
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="span" data-nodeid="2">foo</span> <span class="construct prove" data-nodeid="5"><span class="keyword" data-nodeid="6">PROVE</span> bar</span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_construct_construct():
    compare_have_want(
        have="""
        :pick:foo :: :st: bar ::
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="construct pick assumption" data-nodeid="2"><span class="keyword" data-nodeid="3">PICK</span> foo</span> <span class="construct st assumption" data-nodeid="7"><span class="keyword" data-nodeid="8">SUCH THAT</span> bar</span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_math_math():
    compare_have_want(
        have="""
        $2$ $2$
        """,
        want=r"""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="math" data-nodeid="2">\(2\)</span> <span class="math" data-nodeid="5">\(2\)</span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )
